"""GitLab issue fetching functions"""

import hashlib
from typing import Any

from gitlab import Gitlab

from ._helpers import convert_to_datetime, sort_assignees
from ._types import IssueItem


def _import_gitlab_issues(issues: list[Any], myuser: str, instance_id: str) -> list[IssueItem]:
    """Create a list of IssueItem from the GitLab API results"""
    issueitems: list[IssueItem] = []
    for issue in issues:
        d = IssueItem()
        d.import_values(
            assignee_users=sort_assignees(
                [u["username"] for u in issue.assignees if issue.assignees], myuser
            ),
            due_date=issue.due_date if hasattr(issue, "due_date") else "",
            epic_title=(
                issue.epic["title"] if hasattr(issue, "epic") and issue.epic is not None else ""
            ),
            labels=issue.labels,
            milestone_title=issue.milestone["title"] if issue.milestone else "",
            pull=hasattr(issue, "merge_status"),
            ref=issue.references["full"],
            service="gitlab",
            title=issue.title,
            uid=f"gitlab-{instance_id}-{issue.id}",
            updated_at=convert_to_datetime(issue.updated_at),
            web_url=issue.web_url,
        )
        d.fill_remaining_fields()
        issueitems.append(d)

    return issueitems


def gitlab_get_issues(gitlab: Gitlab) -> list[IssueItem]:
    """Get all issues assigned to authenticated user"""
    issues: list[IssueItem] = []
    myuser: str = gitlab.user.username  # type: ignore
    # Create a unique enough id for the GitLab instance in case we have more
    # than one. Avoids issue id collisions
    instance_id = hashlib.md5(gitlab.url.encode()).hexdigest()[:6]

    # See https://docs.gitlab.com/ee/api/issues.html
    assigned_issues = gitlab.issues.list(
        assignee_username=myuser, state="opened", scope="all", get_all=True
    )
    # See https://docs.gitlab.com/ee/api/merge_requests.html
    merge_requests_assigned = gitlab.mergerequests.list(
        assignee_username=myuser, state="opened", scope="all", get_all=True
    )
    merge_requests_reviews = gitlab.mergerequests.list(
        reviewer_username=myuser, state="opened", scope="all", get_all=True
    )

    issues.extend(
        _import_gitlab_issues(issues=assigned_issues, myuser=myuser, instance_id=instance_id)
    )
    issues.extend(
        _import_gitlab_issues(
            issues=merge_requests_assigned, myuser=myuser, instance_id=instance_id
        )
    )
    issues.extend(
        _import_gitlab_issues(issues=merge_requests_reviews, myuser=myuser, instance_id=instance_id)
    )

    return issues



"""Functions for dealing with a the private tasks repo"""

import logging

from flask import current_app
from github import AuthenticatedUser, Github
from gitlab import Gitlab


def private_tasks_repo_get_gitlab_labels(gitlab: Gitlab) -> dict[str, str]:
    """Get all labels from a GitLab repository"""
    private_tasks_repo = current_app.config["private_tasks_repo"]["repo"]

    all_labels = gitlab.projects.get(private_tasks_repo).labels.list(get_all=True)

    # Return dict of label name and label color
    return {label.name: label.color for label in all_labels}


def private_tasks_repo_get_github_labels(github: Github) -> dict[str, str]:
    """Get all labels from a GitHub repository"""
    private_tasks_repo = current_app.config["private_tasks_repo"]["repo"]

    all_labels = github.get_repo(private_tasks_repo).get_labels()

    # Return dict of label name and label color
    return {label.name: f"#{label.color}" for label in all_labels}


def private_tasks_repo_create_gitlab_issue(gitlab: Gitlab, title: str, labels: list[str]) -> str:
    """Create a new issue in the private tasks repository (GitLab). Returns the
    web URL of the new issue"""
    myuser_id = gitlab.user.id  # type: ignore

    private_tasks_repo = current_app.config["private_tasks_repo"]["repo"]

    # Create issue
    result = gitlab.projects.get(private_tasks_repo).issues.create(
        {"title": title, "labels": labels, "assignee_id": myuser_id}
    )

    logging.debug("Created issue in repository '%s': %s", private_tasks_repo, result.web_url)

    return result.web_url


def private_tasks_repo_create_github_issue(github: Github, title: str, labels: list[str]) -> str:
    """Create a new issue in the private tasks repository (GitHub). Returns the
    web URL of the new issue"""
    private_tasks_repo = current_app.config["private_tasks_repo"]["repo"]
    myuser: AuthenticatedUser.AuthenticatedUser = github.get_user()  # type: ignore

    # Create issue
    result = github.get_repo(private_tasks_repo).create_issue(
        title=title, labels=labels, assignee=myuser.login
    )

    logging.debug("Created issue in repository '%s': %s", private_tasks_repo, result.html_url)

    return result.html_url



"""Main"""

from datetime import datetime

from flask import Blueprint, current_app, flash, redirect, render_template, request
from werkzeug.wrappers import Response

from ._cache import add_to_seen_issues, get_cache_status
from ._views import (
    get_issues_and_stats,
    private_tasks_repo_create_issue,
    private_tasks_repo_get_labels,
    refresh_issues_cache,
    set_ranking,
    set_todolist,
)

main = Blueprint("main", __name__)


@main.route("/", methods=["GET"])
def index() -> str:
    """Index Page"""

    issue_filter = request.args.get("filter", None)

    # Find out whether current cache timer is still valid
    cache = get_cache_status(
        cache_timer=current_app.config["current_cache_timer"],
        timeout_seconds=current_app.config["cache_timeout_seconds"],
    )
    # Reset cache timer to now
    if not cache:
        current_app.config["current_cache_timer"] = datetime.now()

    try:
        issues, stats, new_issues = get_issues_and_stats(cache=cache, issue_filter=issue_filter)
    except Exception as ex:  # pylint: disable=broad-except
        template = "An exception of type {0} occurred. Arguments:\n{1!r}"
        message = template.format(type(ex).__name__, ex.args)
        flash(message, "error")
        return render_template("error.html")

    # Find out if private tasks repo is configured
    private_private_tasks_repo_configured = current_app.config.get("private_tasks_repo", None)

    return render_template(
        "index.html",
        issues=issues,
        stats=stats,
        new_issues=new_issues,
        private_private_tasks_repo_configured=private_private_tasks_repo_configured,
        display_cfg=current_app.config["display"],
        issue_filter=issue_filter,
    )


@main.route("/ranking", methods=["GET"])
def ranking() -> Response:
    """Set ranking"""

    issue = request.args.get("issue", "")
    rank_new = request.args.get("rank", "")

    # Set ranking
    set_ranking(issue=issue, rank=rank_new)
    # When ranking an issue, it also makes the issue be marked as seen
    add_to_seen_issues(new_unseen_issues=[issue])

    return redirect(request.referrer)


@main.route("/todolist", methods=["GET"])
def todolist() -> Response:
    """Add or remove issue from todolist"""

    issue = request.args.get("issue", "")
    state = request.args.get("state", "")

    # Set ranking
    set_todolist(issue=issue, state=state)
    # When ranking an issue, it also makes the issue be marked as seen
    add_to_seen_issues(new_unseen_issues=[issue])

    return redirect(request.referrer)


@main.route("/reload", methods=["GET"])
def reload() -> Response:
    """Reload all issues and break cache"""

    refresh_issues_cache()

    return redirect("/")


@main.route("/mark-as-seen", methods=["GET"])
def mark_as_seen() -> Response:
    """Mark one or all issues as seen"""

    issues = request.args.get("issues", "").split(",")

    add_to_seen_issues(new_unseen_issues=issues)

    return redirect("/")


@main.route("/new", methods=["GET"])
def new_form() -> str:
    """Page form to create new issues"""

    labels = private_tasks_repo_get_labels()

    return render_template(
        "new.html",
        labels=labels,
        colored_labels=current_app.config["private_tasks_repo"]["colored_labels"],
    )


@main.route("/new", methods=["POST"])
def new_create() -> Response:
    """Create a new issue"""

    # Upon cancel request, also refresh cache
    if request.form.get("cancel"):
        refresh_issues_cache()
        return redirect("/")

    # Get relevant form data
    title: str = request.form.get("issue_title", "")
    labels: list[str] = request.form.getlist("labels")

    # Catch potential empty title
    if not title:
        flash("Issue title cannot be empty", "error")
        return redirect("/new")

    # Create new issue
    new_url = private_tasks_repo_create_issue(title=title, labels=labels)
    flash(f"New issue created: <a href='{new_url}' target='_blank'>{new_url}</a>", "success")

    # If user wants back to overview, refresh cache before to also get the newly created issue
    if request.form.get("submit_and_index"):
        refresh_issues_cache()
        return redirect("/")

    return redirect("/new")



"""Module to parse Microsoft Planner JSON export files."""

import json
import logging

from ._helpers import convert_to_datetime
from ._types import IssueItem


class MSPlannerFile:  # pylint: disable=too-few-public-methods
    """Class to parse Microsoft Planner JSON export files."""

    def __init__(self, file: str) -> None:

        try:
            with open(file, "r", encoding="UTF-8") as f:
                self.data = json.load(f)
        except FileNotFoundError as err:
            raise FileNotFoundError(f"Could not find MS Planner export file '{file}'") from err
        except json.JSONDecodeError as err:
            raise ValueError(f"Could not parse MS Planner export file '{file}'") from err


def _import_msplannerfile_issues(issues: list[dict]) -> list[IssueItem]:
    """Create a list of IssueItem from the MS Planner file"""
    issueitems: list[IssueItem] = []
    for issue in issues:
        if issue.get("completedDateTime", False):
            logging.debug("Skipping completed MS Planner task '%s'", issue.get("title", ""))
            continue
        d = IssueItem()
        d.import_values(
            assignee_users="",
            due_date=issue.get("dueDateTime", "")[:10],
            epic_title="",
            labels=[],
            milestone_title="",
            pull="",
            ref="",
            service="msplanner",
            title=issue.get("title", ""),
            uid=f"msplanner-{issue.get('id', '')}",
            updated_at=convert_to_datetime(issue.get("createdDateTime", "")),
            web_url=(
                "https://planner.cloud.microsoft/webui/plan/"
                f"{issue.get('planId', '')}/view/board/task/{issue.get('id', '')}"
            ),
        )
        d.fill_remaining_fields()
        issueitems.append(d)

    return issueitems


def msplannerfile_get_issues(msplannerfile: MSPlannerFile) -> list[IssueItem]:
    """Get all issues assigned to authenticated user"""
    issues: list[IssueItem] = []

    issues.extend(_import_msplannerfile_issues(issues=msplannerfile.data))

    return issues



"""Cache functions"""

import json
import logging
from datetime import datetime, timedelta, timezone
from os.path import join

from platformdirs import user_cache_dir

from ._issues import IssueItem


def _read_cache_file(
    filename: str, instance: str = "list"
) -> list[dict] | dict[str, dict[str, int]]:
    """
    Return a JSON file from the cache directory.

    Args:
        filename (str): The name of the cache file to read.

        instance (str): The type of the expected return value. If "dict", an
        empty dictionary is returned in case of errors. Otherwise, an empty list
        is returned.

    Returns:
        list | dict: The content of the JSON file as a list of dictionaries or a
        dictionary of dictionaries with integer values. Returns an empty list or
        dictionary if the file is not found or cannot be read.
    """
    cache_file = join(user_cache_dir("todo-merger", ensure_exists=True), filename)
    logging.debug("Reading cache file %s", cache_file)

    empty_return: list | dict = {} if instance == "dict" else []

    logging.debug("Reading cache file %s", cache_file)
    try:
        with open(cache_file, mode="r", encoding="UTF-8") as jsonfile:
            return json.load(jsonfile)

    except json.decoder.JSONDecodeError:
        logging.error(
            "Cannot read JSON file %s. Please check its syntax or delete it. "
            "Will ignore any cache.",
            cache_file,
        )
        return empty_return

    except FileNotFoundError:
        logging.debug(
            "Cache file '%s' has not been found. Initializing a new empty one.",
            cache_file,
        )
        return empty_return


def _write_cache_file(filename: str, content: dict | list) -> None:
    """Write a JSON file to the cache directry"""
    cache_file = join(user_cache_dir("todo-merger", ensure_exists=True), filename)

    logging.debug("Writing cache file %s", cache_file)
    with open(cache_file, mode="w", encoding="UTF-8") as jsonfile:
        json.dump(content, jsonfile, indent=2, default=str)


def read_issues_cache() -> list[IssueItem]:
    """Return the current issue cache, or initialize empty one if none present"""
    issues_cache: list[dict] = _read_cache_file(filename="issues.json")  # type: ignore

    if issues_cache:
        # Convert to list of IssueItem
        list_of_dataclasses = []
        for element in issues_cache:
            list_of_dataclasses.append(IssueItem(**element))
        return list_of_dataclasses

    # Initialize empty issues cache
    write_issues_cache(issues=[])
    return []


def write_issues_cache(issues: list[IssueItem]) -> None:
    """Write issues cache file"""
    issues_as_dict = [issue.convert_to_dict() for issue in issues]

    _write_cache_file(filename="issues.json", content=issues_as_dict)


def get_cache_status(cache_timer: None | datetime, timeout_seconds: int) -> bool:
    """Find out whether the cache is still valid. Returns False if it must be
    refreshed"""

    if cache_timer is None:
        logging.debug("No cache timer set before, or manually refreshed")
        return False

    # Get difference between now and start of cache timer
    cache_diff = datetime.now() - cache_timer
    logging.debug("Current cache time difference: %s", cache_diff)
    if cache_diff > timedelta(seconds=timeout_seconds):
        logging.info("Cache older than defined. Requesting all issues anew")
        # Mark that cache shall be disregarded
        return False

    logging.debug("Cache is still considered to be valid")
    return True


def get_unseen_issues(issues: list[IssueItem]) -> dict[str, str]:
    """Return a list of issue IDs that haven't been seen before"""
    # Read seen file
    seen_issues_cached: dict[str, dict[str, int]] = _read_cache_file(
        filename="seen-issues.json", instance="dict"
    )  # type: ignore

    unseen_issues = {}

    for issue in issues:
        if issue.uid not in seen_issues_cached:
            logging.debug("Issue %s hasn't been seen before", issue.uid)
            unseen_issues[issue.uid] = issue.title

    return unseen_issues


def add_to_seen_issues(new_unseen_issues: list[str]) -> None:
    """Add one or multiple issues to the seen issues list"""

    # Read seen file
    seen_issues_cached: dict[str, dict[str, int]] = _read_cache_file(
        filename="seen-issues.json", instance="dict"
    )  # type: ignore

    # Extend seen issues with new list
    for new_issue in new_unseen_issues:
        logging.debug("Marking issue %s as seen", new_issue)
        seen_issues_cached[new_issue] = {"first_seen": int(datetime.now(timezone.utc).timestamp())}

    # Update file
    _write_cache_file(filename="seen-issues.json", content=seen_issues_cached)


def update_last_seen() -> None:
    """Update the last_seen flag of an issue in the cache"""
    issues_cache: list[dict] = _read_cache_file(filename="issues.json")  # type: ignore
    seen_issues_cached: dict[str, dict[str, int]] = _read_cache_file(
        filename="seen-issues.json", instance="dict"
    )  # type: ignore

    logging.debug("Updating last_seen flag for all %s issues in cache", len(issues_cache))
    for issue in issues_cache:
        uid = issue.get("uid", "")
        if uid in seen_issues_cached:
            seen_issues_cached[uid]["last_seen"] = int(datetime.now(timezone.utc).timestamp())
        else:
            logging.debug("Issue %s not found in seen issues cache", uid)

    _write_cache_file(filename="seen-issues.json", content=seen_issues_cached)



"""Shared type definitions to avoid circular imports"""

from dataclasses import asdict, dataclass, field
from datetime import datetime

from ._helpers import convert_to_datetime, time_ago


@dataclass
class IssueItem:  # pylint: disable=too-many-instance-attributes
    """Dataclass holding a single issue"""

    assignee_users: list = field(default_factory=list)
    due_date: str = ""
    epic_title: str = ""
    labels: list = field(default_factory=list)
    milestone_title: str = ""
    pull: bool = False
    rank: int = 5  # Will be updated with ISSUE_RANKING_TABLE["normal"]
    ref: str = ""
    service: str = ""
    title: str = ""
    todolist: bool = False
    uid: str = ""
    updated_at_display: str = ""
    updated_at: datetime = field(default_factory=datetime.now)
    web_url: str = ""

    def import_values(self, **kwargs):
        """Import data from a dict"""
        for attr, value in kwargs.items():
            setattr(self, attr, value)

    def fill_remaining_fields(self):
        """Fill remaining fields that have not been imported directly and which
        are solely derived from attribute values"""

        # updated_at_display
        self.updated_at_display = time_ago(convert_to_datetime(self.updated_at))

    def convert_to_dict(self):
        """Return the current dataclass as dict"""
        return asdict(self)


@dataclass
class IssuesStats:  # pylint: disable=too-many-instance-attributes
    """Dataclass holding a stats about all issues"""

    total: int = 0
    gitlab: int = 0
    github: int = 0
    msplanner: int = 0
    pulls: int = 0
    issues: int = 0
    due_dates_total: int = 0
    milestones_total: int = 0
    epics_total: int = 0



"""App config file handling"""

import json
import logging
import sys
from os import makedirs
from os.path import dirname, join

import toml
from platformdirs import user_config_dir

DEFAULT_APP_CONFIG = """# App configuration for ToDo Merger

[cache]
timeout_seconds = 600

[services.github-com]
service = "github"
token = ""

[services.gitlab-com]
service = "gitlab"
token = ""
url = "https://gitlab.com"

# Private tasks repository configuration
# [private-tasks-repo]
# service = "github-com"
# repo = "my-username/todos"
# colored_labels = true

[display]
show_assignees = true
show_due_date = true
show_epic = true
show_labels = true
show_milestone = true
show_ref = true
show_service = true
show_type = true
show_updated_at = true
show_web_url = true
"""


def _initialize_config_file(configfile: str) -> dict:
    """Create a new app configuration file with default values"""

    # Create directory in case it does not exist
    makedirs(dirname(configfile), exist_ok=True)

    # Write the default config in TOML format to a new file
    with open(configfile, mode="w", encoding="UTF-8") as tomlfile:
        tomlfile.write(DEFAULT_APP_CONFIG)

    return toml.loads(DEFAULT_APP_CONFIG)


def _read_app_config_file(config_file: str) -> dict:
    """Read full app configuration"""
    try:
        with open(config_file, mode="r", encoding="UTF-8") as tomlfile:
            app_config = toml.load(tomlfile)

    except FileNotFoundError:
        logging.warning(
            "App configuration file '%s' has not been found. Initializing a new empty one.",
            config_file,
        )
        app_config = _initialize_config_file(config_file)

    except toml.decoder.TomlDecodeError:
        logging.error("Error reading configuration file '%s'. Check the syntax!", config_file)
        sys.exit(1)

    return app_config


def default_config_file_path() -> str:
    """Define the path of the config file"""
    return join(user_config_dir("todo-merger", ensure_exists=True), "config.toml")


def get_app_config(config_file: str, key: str = "", warn_on_missing_key: bool = True) -> dict:
    """Return a specific section from the app configuration, or the whole config"""
    logging.debug("Reading app configuration file %s", config_file)

    if not config_file:
        config_file = default_config_file_path()

    if key:
        try:
            return _read_app_config_file(config_file)[key]
        except KeyError as e:
            if warn_on_missing_key:
                logging.warning("Key %s not found in configuration: %s", key, e)
            return {}

    return _read_app_config_file(config_file)


def read_issues_config() -> dict:
    """Return the issues configuration"""

    config_file = join(user_config_dir("todo-merger", ensure_exists=True), "issues-config.json")

    logging.debug("Reading issues configuration file %s", config_file)
    try:
        with open(config_file, mode="r", encoding="UTF-8") as jsonfile:
            return json.load(jsonfile)

    except json.decoder.JSONDecodeError:
        logging.error(
            "Cannot read JSON file %s. Please check its syntax or delete it. "
            "Will ignore any issues configuration.",
            config_file,
        )
        return {}

    except FileNotFoundError:
        logging.debug(
            "Issues configuration file '%s' has not been found. Initializing a new empty one.",
            config_file,
        )
        default_issues_config: dict = {}
        write_issues_config(issues_config=default_issues_config)

        return default_issues_config


def write_issues_config(issues_config: dict) -> None:
    """Write issues configuration file"""

    config_file = join(user_config_dir("todo-merger", ensure_exists=True), "issues-config.json")

    logging.debug("Writing issues configuration file %s", config_file)
    with open(config_file, mode="w", encoding="UTF-8") as jsonfile:
        json.dump(issues_config, jsonfile)



"""View functions, removing complexity from main.py"""

import logging

from flask import current_app

from ._cache import (
    get_unseen_issues,
    read_issues_cache,
    update_last_seen,
    write_issues_cache,
)
from ._config import read_issues_config, write_issues_config
from ._issues import (
    ISSUE_RANKING_TABLE,
    IssueItem,
    IssuesStats,
    apply_issue_filter,
    apply_user_issue_config,
    get_all_issues,
    get_issues_stats,
    prioritize_issues,
)
from ._private_tasks import (
    private_tasks_repo_create_github_issue,
    private_tasks_repo_create_gitlab_issue,
    private_tasks_repo_get_github_labels,
    private_tasks_repo_get_gitlab_labels,
)


def get_issues_and_stats(
    cache: bool, issue_filter: str | None
) -> tuple[list[IssueItem], IssuesStats, dict[str, str]]:
    """Functions to view all issues. Returns: list of IssueItem, a IssueStats
    object, and list of issue IDs"""
    # Get issues (either cache or online)
    if cache:
        issues = read_issues_cache()
    else:
        # Get all issues from the services
        issues = get_all_issues()
        # Update cache file
        write_issues_cache(issues=issues)
        # Update last_seen flag in seen issues cache
        update_last_seen()
    # Get previously unseen issues
    new_issues = get_unseen_issues(issues=issues)
    # Default prioritization
    issues = prioritize_issues(issues)
    # Issues custom config (ranking)
    config = read_issues_config()
    issues = apply_user_issue_config(issues=issues, issue_config_dict=config)
    # Apply issue filter
    issues = apply_issue_filter(issues=issues, issue_filter=issue_filter)
    # Stats
    stats = get_issues_stats(issues)

    return issues, stats, new_issues


def set_ranking(issue: str, rank: str) -> None:
    """Set new ranking of individual issue inside of the issues configuration file"""
    rank_int = ISSUE_RANKING_TABLE.get(rank, ISSUE_RANKING_TABLE["normal"])
    config: dict[str, dict[str, int | bool]] = read_issues_config()

    # Catch undefined issues
    if not issue:
        return

    # Create new issue entry if it does not exist
    if issue not in config:
        config[issue] = {}

    # Check if new ranking is the same as old -> reset to default
    if issue in config and config.get(issue, {}).get("rank") == rank_int:
        logging.info("Resetting ranking for issue '%s'", issue)
        config[issue].pop("rank", None)
    # Setting new ranking value
    else:
        logging.info("Setting rank of issue '%s' to %s (%s)", issue, rank, rank_int)
        config[issue]["rank"] = rank_int

    # Update config file
    write_issues_config(issues_config=config)


def set_todolist(issue: str, state: str | bool) -> None:
    """Add or remove an issue from the personal todo list"""
    config = read_issues_config()

    # Convert state to bool if str
    if isinstance(state, str):
        state = state.lower() in ("yes", "true", "t", "1")

    # Catch undefined issues
    if not issue:
        return

    # Create new issue entry if it does not exist
    if issue not in config:
        config[issue] = {}

    if state:
        logging.info("Adding issue '%s' to the todo list", issue)
    else:
        logging.info("Removing issue '%s' from the todo list", issue)
    config[issue]["todolist"] = state

    # Update config file
    write_issues_config(issues_config=config)


def refresh_issues_cache() -> None:
    """Refresh the cache of issues"""
    current_app.config["current_cache_timer"] = None


def private_tasks_repo_get_labels() -> dict[str, str]:
    """Get all labels from the private tasks repository"""
    service, login = (
        current_app.config["private_tasks_repo"]["service"],
        current_app.config["private_tasks_repo"]["login"],
    )

    if service == "gitlab":
        return private_tasks_repo_get_gitlab_labels(gitlab=login)

    return private_tasks_repo_get_github_labels(github=login)


def private_tasks_repo_create_issue(title: str, labels: list[str]) -> str:
    """Create a new issue in the private tasks repository. Returns the web URL
    of the new issue"""
    service, login = (
        current_app.config["private_tasks_repo"]["service"],
        current_app.config["private_tasks_repo"]["login"],
    )

    if service == "gitlab":
        return private_tasks_repo_create_gitlab_issue(gitlab=login, title=title, labels=labels)

    return private_tasks_repo_create_github_issue(github=login, title=title, labels=labels)



"""ToDo Merger: Provide an overview of your assigned issues on GitLab and GitLab"""

import argparse
import logging
import signal
import sys
import uuid
from importlib.metadata import version
from os import kill, path

from flask import Flask
from github import Github
from gitlab import Gitlab
from platformdirs import user_log_dir, user_runtime_dir
from sassutils.wsgi import SassMiddleware

from ._msplanner import MSPlannerFile

try:
    import daemon
    import daemon.pidfile
# pwd does not exist on Windows, we cannot daemonize there
except (ModuleNotFoundError, ImportError):
    daemon = None

from ._auth import github_login, gitlab_login
from ._config import default_config_file_path, get_app_config

LOGFILE = path.join(user_log_dir("todo-merger", ensure_exists=True), "todo-merger.log")
PIDFILE = path.join(user_runtime_dir("todo-merger", ensure_exists=True), "todo-merger.pid")

__version__ = version("todo-merger")

parser = argparse.ArgumentParser(
    description=__doc__, formatter_class=argparse.ArgumentDefaultsHelpFormatter
)
subparsers = parser.add_subparsers(dest="command", help="Special commands")

parser.add_argument(
    "-c", "--config-file", help="Path to the app config file ", default=default_config_file_path()
)
parser.add_argument(
    "-d",
    "--daemon",
    help="Start the app in the background and log to a file",
    action="store_true",
)
parser.add_argument(
    "-l", "--logfile", help="Path to the log file (in daemon mode)", default=LOGFILE
)
parser.add_argument(
    "-p", "--pidfile", help="Path to the PID file (in daemon mode)", default=PIDFILE
)
parser.add_argument("--port", help="Port the application runs on", type=int, default=8636)
parser.add_argument(
    "-v",
    "--verbose",
    help="Activate INFO logging",
    action="store_true",
)
parser.add_argument(
    "-vv",
    "--debug",
    help="Activate DEBUG logging (even more than INFO)",
    action="store_true",
)
parser.add_argument("--version", action="version", version="%(prog)s " + __version__)

# Special commands
parser_stop = subparsers.add_parser(
    "stop",
    help="Stop the running todo-merger background instance",
)


def configure_logger(args) -> logging.Logger:
    """Set logging options"""
    log = logging.getLogger()
    logging.basicConfig(
        encoding="utf-8",
        format="%(levelname)s in %(module)s: %(message)s",
        level=(logging.DEBUG if args.debug else logging.INFO if args.verbose else logging.WARNING),
        handlers=[logging.StreamHandler()],
    )

    return log


def load_app_services_config(
    config_file: str, section: str = "services"
) -> dict[str, tuple[str, Github | Gitlab | MSPlannerFile]]:
    """Load the app config, handle service logins, and return objects"""

    app_config: dict[str, dict[str, str]] = get_app_config(config_file, section)
    service_objects: dict[str, tuple[str, Github | Gitlab | MSPlannerFile]] = {}

    for name, cfg in app_config.items():
        service = cfg.get("service", "")
        url = cfg.get("url", "")
        token = cfg.get("token", "")

        if not service:
            logging.critical(
                "The config section %s has no 'service' defined, e.g. 'github' or 'gitlab'", name
            )
            sys.exit(1)

        if service == "gitlab" and not url:
            logging.critical(
                "The config section %s is a gitlab service but has no 'url' defined", name
            )
            sys.exit(1)

        if name in service_objects:
            logging.critical(
                "You have used the section name %s more than once. Please make them unique", name
            )
            sys.exit(1)

        loginobj: Github | Gitlab | MSPlannerFile
        if service == "github":
            loginobj = github_login(token)
        elif service == "gitlab":
            loginobj = gitlab_login(token, url)  # type: ignore
        elif service == "msplanner-file":
            loginobj = MSPlannerFile(cfg.get("file", ""))
        else:
            logging.critical("The config section %s contains an unknown 'service'", name)
            sys.exit(1)

        service_objects[name] = (service, loginobj)

    return service_objects


# pylint: disable=import-outside-toplevel
def create_app(config_file: str):
    """Create Flask App"""

    # Initiate Flask app
    app = Flask(__name__)
    # Reload templates
    app.jinja_env.auto_reload = True
    app.jinja_env.lstrip_blocks = False
    app.jinja_env.globals["app_version"] = __version__
    app.config["TEMPLATES_AUTO_RELOAD"] = True

    # Set werkzeug logging level to the global logging level
    logging.getLogger("werkzeug").setLevel(logging.root.level)

    # Settings for SCSS conversion
    app.wsgi_app = SassMiddleware(  # type: ignore
        app.wsgi_app,
        {
            "todo_merger": {
                "sass_path": "static/sass",
                "css_path": "static/css",
                "wsgi_path": "/static/css",
                "strip_extension": False,
            }
        },
    )

    # Set a secret key for the session
    app.secret_key = str(uuid.uuid4().hex)

    # Load app config and login to services (e.g. GitHub and GitLab)
    app.config["services"] = {}
    for name, service in load_app_services_config(config_file).items():
        app.config["services"][name] = service

    # Initiate cache timer
    app.config["current_cache_timer"] = None
    app.config["cache_timeout_seconds"] = get_app_config(config_file, "cache").get(
        "timeout_seconds", 600
    )

    # Load display config and set default values
    app.config["display"] = get_app_config(config_file, "display")
    for display_cfg in (
        "show_assignees",
        "show_due_date",
        "show_epic",
        "show_labels",
        "show_milestone",
        "show_ref",
        "show_service",
        "show_type",
        "show_updated_at",
        "show_web_url",
    ):
        app.config["display"][display_cfg] = app.config["display"].get(display_cfg, True)

    # Get private-tasks-repo config
    if private_tasks_repo_config := get_app_config(
        config_file, "private-tasks-repo", warn_on_missing_key=False
    ):
        app.config["private_tasks_repo"] = private_tasks_repo_config
        # Find the GitHub/GitLab service object that is configured for the private tasks repo
        try:
            (
                app.config["private_tasks_repo"]["service"],
                app.config["private_tasks_repo"]["login"],
            ) = app.config["services"][app.config["private_tasks_repo"]["service"]]
        except KeyError:
            logging.critical(
                "The 'private-tasks-repo' section in the config file refers to "
                "a service that is not defined"
            )
            sys.exit(1)
    else:
        logging.info(
            "No 'private-tasks-repo' section found in config file. Disabling this functionality"
        )
        app.config["private_tasks_repo"] = None

    # Print app config in DEBUG
    logging.debug("App config: %s", app.config)

    # blueprint for app
    from .main import main as main_blueprint  # pylint: disable=import-error

    app.register_blueprint(main_blueprint)

    return app


def run_server(config_file: str, port: int):
    """Run the Flask server"""
    app = create_app(config_file=config_file)
    app.run(port=port)


def main():
    """Main entry point for running the app."""
    args = parser.parse_args()

    # Configure logger
    logger = configure_logger(args=args)

    # Stop app if requested
    if args.command == "stop":
        try:
            with open(args.pidfile, "r", encoding="utf-8") as pidreader:
                pid = int(pidreader.read())
                print(f"Sending SIGTERM to process {pid}")
                kill(pid, signal.SIGTERM)
                sys.exit()
        except FileNotFoundError:
            sys.exit(
                f"PID file {args.pidfile} does not seem to exist. The app does not seem to run "
                "normally or at all. Check your task manager and process list."
            )

    # Start app
    print(f"ToDo Merger will be available on http://localhost:{args.port}")
    logging.info("Config file: %s", args.config_file)
    logging.info("Log file: %s", args.logfile)
    logging.info("PID file: %s", args.pidfile)
    if args.daemon:
        if daemon is None:
            sys.exit(
                "Daemonizing this app is not possible on your system, e.g. because it's Windows."
            )
        with daemon.DaemonContext(pidfile=daemon.pidfile.TimeoutPIDLockFile(args.pidfile)):
            # Add file logger
            logger.addHandler(logging.FileHandler(args.logfile))

            # Run server
            run_server(config_file=args.config_file, port=args.port)
    else:
        run_server(config_file=args.config_file, port=args.port)



"""Helper functions"""

from datetime import datetime, timezone

from dateutil import parser


def sort_assignees(assignees: list, my_user_name: str) -> str:
    """Provide a human-readable list of assigned users, treating yourself special"""

    # Remove my user name from assignees list, if present
    try:
        assignees.remove(my_user_name)
    except ValueError:
        pass

    # If executing user is the only assignee, there is no use in that field
    if not assignees:
        return ""

    return f"{', '.join(['Me'] + assignees)}"


def convert_to_datetime(timestamp: str | datetime) -> datetime:
    """
    Convert a timestamp string or datetime to a timezone-aware datetime object.

    Args:
        timestamp (str | datetime): The timestamp string to convert, or the
        datetime object to pass through

    Returns:
        datetime: A timezone-aware datetime object in UTC.
    """
    # Handle if timestamp is already datetime object
    if isinstance(timestamp, datetime):
        dt = timestamp
    # Convert str to datetime
    else:
        try:
            # Attempt to parse with dateutil parser
            dt = parser.isoparse(timestamp)

        except ValueError as exc:
            raise ValueError(f"Unrecognized timestamp format: {timestamp}") from exc

    # If the datetime object is naive (no timezone), assume UTC
    if dt.tzinfo is None:
        dt = dt.replace(tzinfo=timezone.utc)

    return dt.astimezone(timezone.utc)


def time_ago(dt):
    """Convert a datetime to a human-readable 'time ago' format"""
    now = datetime.now(timezone.utc)
    diff = now - dt

    if diff.days >= 365:
        years = diff.days // 365
        display = f"{years} year{'s' if years > 1 else ''} ago"
    elif diff.days >= 30:
        months = diff.days // 30
        display = f"{months} month{'s' if months > 1 else ''} ago"
    elif diff.days >= 7:
        weeks = diff.days // 7
        display = f"{weeks} week{'s' if weeks > 1 else ''} ago"
    elif diff.days > 0:
        display = f"{diff.days} day{'s' if diff.days > 1 else ''} ago"
    elif diff.seconds >= 3600:
        hours = diff.seconds // 3600
        display = f"{hours} hour{'s' if hours > 1 else ''} ago"
    elif diff.seconds >= 60:
        minutes = diff.seconds // 60
        display = f"{minutes} minute{'s' if minutes > 1 else ''} ago"
    else:
        display = "Just now"

    return display



"""Handling issues which come in from different sources"""

import logging
from dataclasses import fields
from datetime import datetime

from flask import current_app

from ._types import IssueItem, IssuesStats

ISSUE_RANKING_TABLE = {"pin": -1, "high": 1, "normal": 5, "low": 99}


# ----------------------------------------
# ISSUES FETCHING FROM ALL SERVICES
# ----------------------------------------


def get_all_issues() -> list[IssueItem]:
    """Get all issues from the supported services"""
    # Import here to avoid circular dependency
    # pylint: disable=import-outside-toplevel
    from ._github import github_get_issues
    from ._gitlab import gitlab_get_issues
    from ._msplanner import msplannerfile_get_issues

    issues: list[IssueItem] = []
    for name, service in current_app.config["services"].items():
        if service[0] == "github":
            logging.info("Getting assigned GitHub issues for %s", name)
            issues.extend(github_get_issues(service[1]))
        elif service[0] == "gitlab":
            logging.info("Getting assigned GitLab issues for %s", name)
            issues.extend(gitlab_get_issues(service[1]))
        elif service[0] == "msplanner-file":
            logging.info("Getting assigned MS Planner tasks for %s", name)
            issues.extend(msplannerfile_get_issues(service[1]))
        else:
            print(f"Service {service[0]} not supported for fetching issues")

    return issues


# ----------------------------------------
# ISSUE PRIORIZATION AND FILTERING
# ----------------------------------------


def _replace_none_with_empty_string(obj: IssueItem) -> IssueItem:
    """Replace None values of a dataclass with an empty string. Makes sorting
    easier"""
    for f in fields(obj):
        value = getattr(obj, f.name)
        if value is None:
            setattr(obj, f.name, "")

    return obj


def prioritize_issues(
    issues: list[IssueItem], sort_by: list[tuple[str, bool]] | None = None
) -> list[IssueItem]:
    """
    Sorts the list of IssueItem objects based on multiple criteria.

    :param issues: List of IssueItem objects to sort.

    :param sort_by: List of tuples where each tuple contains:
                    - field name to sort by as a string
                    - a boolean indicating whether to sort in reverse order
                      (True for descending, False for ascending)

    :return: Sorted list of IssueItem objects.
    """
    if sort_by is None:
        sort_by = [
            ("due_date", False),
            ("milestone_title", True),
            ("epic_title", True),
            ("updated_at", True),
        ]

    logging.info("Sort issues based on %s", sort_by)

    # Replace None with empty string in all tasks
    issues = [_replace_none_with_empty_string(task) for task in issues]

    def sort_key(issue: IssueItem) -> tuple:
        # Create a tuple of the field values to sort by, considering the reverse order
        key: list[tuple[int, str | None]] = []
        for f, reverse in sort_by:
            value: str | datetime = getattr(issue, f)
            # Convert datetime to str
            if isinstance(value, datetime):
                value = value.strftime("%s")
            elif isinstance(value, str):
                value = value.lower()
            is_empty: bool = value == ""
            # Place empty values at the end
            if is_empty:
                key.append((1, None))  # `1` indicates an empty value
            else:
                if reverse:
                    value = "".join(chr(255 - ord(char)) for char in value)
                key.append((0, value))  # `0` indicates a non-empty value
        return tuple(key)

    return sorted(issues, key=sort_key)


def apply_user_issue_config(
    issues: list[IssueItem], issue_config_dict: dict[str, dict[str, int | bool]]
) -> list[IssueItem]:
    """Apply local user configuration to issues"""
    for issue in issues:
        if issue.uid in issue_config_dict:
            issue.rank = issue_config_dict[issue.uid].get("rank", ISSUE_RANKING_TABLE["normal"])
            logging.debug("Applied rank %s to issue %s (%s)", issue.rank, issue.uid, issue.title)
            if issue_config_dict[issue.uid].get("todolist", False):
                logging.debug("Put issue %s on todo list (%s)", issue.uid, issue.title)
                issue.todolist = True

    return issues


def apply_issue_filter(issues: list[IssueItem], issue_filter: str | None) -> list[IssueItem]:
    """Apply issue filter to issues"""
    if not issue_filter:
        logging.debug("No issue filter applied")

    logging.info("Applying issue filter '%s'", issue_filter)

    if issue_filter == "todolist":
        issues = [issue for issue in issues if issue.todolist]

    return issues


# ----------------------------------------
# STATS ABOUT FETCHED ISSUES
# ----------------------------------------


def get_issues_stats(issues: list[IssueItem]) -> IssuesStats:
    """Create some stats about the collected issues"""
    stats = IssuesStats()

    for issue in issues:
        # Total issues
        stats.total += 1
        # Services total
        setattr(stats, issue.service, getattr(stats, issue.service) + 1)
        # Issue/PR counter
        if issue.pull:
            stats.pulls += 1
        else:
            stats.issues += 1
        # Number of due dates, milestones, and epics
        stats.due_dates_total += 1 if issue.due_date else 0
        stats.milestones_total += 1 if issue.milestone_title else 0
        stats.epics_total += 1 if issue.epic_title else 0

    return stats



"""GitHub issue fetching functions"""

from typing import Any
from urllib.parse import urlparse

from github import AuthenticatedUser, Github, Issue, PaginatedList

from ._helpers import convert_to_datetime, sort_assignees
from ._types import IssueItem


def _gh_url_to_ref(url: str):
    """Convert a GitHub issue URL to a ref"""
    url = urlparse(url).path
    url = url.strip("/")

    # Run replacements
    replacements = {"/issues/": "#", "/pull/": "#"}
    for search, replacement in replacements.items():
        url = url.replace(search, replacement)

    return url


def _import_github_issues(
    issues: PaginatedList.PaginatedList[Issue.Issue] | Any, myuser: str
) -> list[IssueItem]:
    """Create a list of IssueItem from the GitHub API results"""
    # pylint: disable=duplicate-code
    issueitems: list[IssueItem] = []
    for issue in issues:
        d = IssueItem()
        d.import_values(
            assignee_users=sort_assignees(
                [u.login for u in issue.assignees if issue.assignees], myuser
            ),
            due_date="",
            epic_title="",
            labels=[label.name for label in issue.labels],
            milestone_title=issue.milestone.title if issue.milestone else "",
            # Ugly fix to make loading of whether it's a PR faster.
            # `issue.pull_request` would trigger another API call
            pull="/pull/" in issue.html_url,
            ref=_gh_url_to_ref(issue.html_url),
            service="github",
            title=issue.title,
            uid=f"github-{issue.id}",
            updated_at=convert_to_datetime(issue.updated_at),
            web_url=issue.html_url,
        )
        d.fill_remaining_fields()
        issueitems.append(d)

    return issueitems


def github_get_issues(github: Github) -> list[IssueItem]:
    """Get all issues assigned to authenticated user"""
    issues: list[IssueItem] = []
    myuser: AuthenticatedUser.AuthenticatedUser = github.get_user()  # type: ignore

    # See https://docs.github.com/en/rest/issues/issues
    assigned_issues = myuser.get_issues()
    # See https://docs.github.com/en/rest/search/search
    review_requests = github.search_issues(
        query=f"is:open is:pr archived:false review-requested:{myuser.login}"
    )

    issues.extend(_import_github_issues(issues=assigned_issues, myuser=myuser.login))
    issues.extend(_import_github_issues(issues=review_requests, myuser=myuser.login))

    return issues



"""Login functions"""

import logging
import sys

from github import Github
from gitlab import Gitlab

from ._config import default_config_file_path


def github_login(token: str) -> Github:
    """Login to GitHub with token"""
    if not token:
        logging.critical(
            "You need to provide a token for GitHub. Add that to your configuration file. "
            "Default: %s",
            default_config_file_path(),
        )
        sys.exit(1)
    g = Github(login_or_token=token)
    logging.info("Logged into GitHub as %s", g.get_user().login)
    return g


def gitlab_login(token: str, url: str = "https://gitlab.com") -> Gitlab:
    """Login to GitHub with token"""
    if not token:
        logging.critical(
            "You need to provide a token for GitLab. Add that to your configuration file. "
            "Default: %s",
            default_config_file_path(),
        )
        sys.exit(1)

    g = Gitlab(url=url, private_token=token)
    # Perform login, populates the user attribute
    g.auth()
    logging.info("Logged into GitLab as %s on %s", g.user.username, url)  # type: ignore
    return g
